__author__ = 'krishvi7'
import os
import sys
import os.path
import re
import datetime
import barcode_count
import simplify
import matricize
import tall_skinny
import cw_qualitysummary

# Contains functions used across the package
class Common():

    # Default values for paramters
    output_prefix = "output"
    barcode_index = "AGCAGAG"
    phred_format = 33

    def __init__(self):
        self.dataType_dict = {"input_files": str,
                              "output_prefix": str,
                              "input_directory": str,
                              "output_directory": str,
                              "phred_format": int,
                              "barcode_index": str,
                              "sample_pattern_regex": str,
                              "known_contaminant": str,
                              "suppress_matrix": bool,
                              "suppress_quality_cw_summary": bool,
                              "suppress_tall_skinny": bool,
                              "library_mode": bool}
        self.range_dict = {"phred_format": [33, 64]}
        self.paramless_dict = {"suppress_quality_cw_summary": True,
                               "suppress_matrix": True,
                               "suppress_tall_skinny": True,
                               "library_mode": True}
        return

    @staticmethod
    def print_error(message):
        print >> sys.stderr, "ERROR: "+message
        sys.exit(1)

    @staticmethod
    def get_timestamp():
        return datetime.datetime.utcnow().strftime('%Y%m%d%H%M%S')

    @staticmethod
    def super_usage():
        print("\nClonTracer package v1.2- ClonTracer package is used to analyze barcode data obtained from ClonTracer"
              " library")


# Contains code to validate a list of input files
class ClonTracerCountBarcodes():

    def __init__(self, delimiter=".", input_files="", input_file_list=list(), output_prefix="output",
                 barcode_index=Common.barcode_index, known_contaminant="", phred_format=Common.phred_format, library_mode=False):
        self.delimiter = delimiter
        self.input_file_list = input_file_list
        self.input_files = input_files
        self.output_prefix = output_prefix
        self.barcode_index = barcode_index
        self.phred_format = phred_format
        self.module = "clonTracer_countBarcodes"
        self.barcode_file = ""
        self.test_mode = False
        self.library_mode = library_mode
        self.known_contaminant = known_contaminant
        self.file_extension_list = [".fastq", ".fq", ".fastq.gz", ".fastq.bz2"]
        self.args_with_params = {"input_files", "output_prefix", "barcode_index", "known_contaminant"}
        self.args_dict = {"input_files": "i", "output_prefix": "o", "phred_format": "q", "barcode_index": "b",
                          "known_contaminant": "k"}
        self.args_dict_paramless = {"library_mode": "l"}
        self.args_dict_all = self.args_dict.copy()
        self.args_dict_all.update(self.args_dict_paramless)
        self.reverse_args_dict_all = {value: key for key, value in self.args_dict_all.iteritems()}
        self.args_description = {"i": "Comma separated list of fastq files. Supports "+",".join(self.file_extension_list),
                                 "o": "Output prefix",
                                 "q": "phred quality format. must be 33 or 64",
                                 "b": "primer index following WS barcode",
                                 "k": "Text file containing known contaminant barcodes",
                                 "l": "Library mode excludes barcode merging step"}
        self.options = ["o", "q", "b", "k", "l"]
        self.args_default = {"o": "output", "b": "AGCAGAG", "q": "33"}

    def usage(self):
        Common.super_usage()
        print("\nUsage:")
        print("-------")
        print("  python clonTracer_countBarcodes.py [options]* -i|--input_files <ip_files>")
        print("\n  -i|--input_files "+self.args_description["i"])
        print("\noptions:")
        print("--------")
        for option in self.options:
            print ("  -"+option+"|--"+self.reverse_args_dict_all[option]+"\t: "+self.args_description[option] +
                   (". Default- "+self.args_default[option] if option in self.args_default.keys() else ""))
        print ("")
        sys.exit(1)

    def check_input_files(self):
        if self.input_files == "":
            Common.print_error("Input file(s) for barcode counting not supplied")

        for input_file in self.input_files.split(","):
            if not os.path.isfile(input_file):
                Common.print_error("Input file "+input_file+" does not exist")

            if not any(input_file.endswith(x) for x in self.file_extension_list):
                Common.print_error("Invalid input "+input_file+", Must be a *"+' ,*'.join(self.file_extension_list))

    def count_barcodes(self):
        self.barcode_file = self.output_prefix+".barcode_verbose.txt"
        print("\nCounting barcodes on sample "+self.input_files+"....")
        barcode_count.barcode_count(self.input_files, self.barcode_file, self.barcode_index, self.known_contaminant,
                                    self.phred_format, self.library_mode)

    def simplify_barcodes(self):
        print("Simplifying barcode from sample "+self.output_prefix+"....")
        simplify.simplify_barcodes(self.barcode_file, self.output_prefix+".barcode.txt", self.output_prefix +
                                   ".cumulative_wealth.txt")


# Contains code to validate an input directory and submits individual samples
class ClonTracerCountBarcodesExperiment():

    def __init__(self):
        self.input_directory = ""
        self.output_directory = Common.output_prefix+"_"+Common.get_timestamp()
        self.suppress_quality_cw_summary = False
        self.library_mode = False
        self.suppress_matrix = False
        self.suppress_tall_skinny = False
        self.barcode_index = Common.barcode_index
        self.known_contaminant = ""
        self.phred_format = Common.phred_format
        self.sample_pattern_regex = ""
        self.test_mode = False
        self.module = "clonTracer_countBarcodes_experiment"
        self.args_dict = {"input_directory": "i", "output_directory": "o", "phred_format": "q",
                          "barcode_index": "b",  "known_contaminant": "k", "sample_pattern_regex": "p"}
        self.args_dict_paramless = {"suppress_quality_cw_summary": "c", "suppress_matrix": "m",
                                    "suppress_tall_skinny": "s", "library_mode": "l"}
        self.args_dict_all = self.args_dict.copy()
        self.args_dict_all.update(self.args_dict_paramless)
        self.reverse_args_dict_all = {value: key for key, value in self.args_dict_all.iteritems()}
        self.args_description = {"i": "Input directory containing the folder structure as described in the "
                                      "documentation",
                                 "o": "Output directory",
                                 "c": "Do not produce the quality and cumulative summary file",
                                 "m": "Do not produce barcodes x samples matrix of barcode fraction",
                                 "s": "Do not produce tall-skinny version of barcode-sample matrix",
                                 "q": "phred quality format. must be 33 or 64",
                                 "b": "primer index following WS barcode",
                                 "k": "Text file containing known contaminant barcodes",
                                 "p": "Sample pattern regex",
                                 "l": "Library mode excludes barcode merging step"}
        self.options = ["o", "c", "m", "s", "q", "b", "k", "p", "l"]
        self.args_default = {"o": "./output_<datetimevalue>", "b": "AGCAGAG", "q": "33"}

    def usage(self):
        Common.super_usage()
        print("\nUsage:")
        print("-------")
        print("  python clonTracer_countBarcodes_experiment.py [options]* -i|--input_directory <ip_dir>")
        print("\n  -i|--input_directory: "+self.args_description["i"])
        print("\noptions:")
        print("--------")
        for option in self.options:
            print ("  -"+option+"|--"+self.reverse_args_dict_all[option]+"\t: "+self.args_description[option] +
                   (". Default- "+self.args_default[option] if option in self.args_default.keys() else ""))

        print("")
        sys.exit(1)

    # Checks the existence of input directory and prints appropriate messages
    def check_input_directory(self):

        print("Checking input directory")
        if self.input_directory == "":
            Common.print_error("Input directory for barcode counting not supplied")

        if not os.path.isdir(self.input_directory):
            Common.print_error("Input directory "+self.input_directory + " does not exist")

        print("Input directory "+self.input_directory+" found")

    # Checks the existence of the output directory
    def check_output_directory(self):
        if not os.path.exists(self.output_directory):
            if not self.test_mode:
                os.makedirs(self.output_directory)
        print("Output would be stored in "+self.output_directory)

    # Checks the existence of the output directory
    def check_known_contaminant_file(self):
        if self.known_contaminant != "":
            if not os.path.exists(self.known_contaminant):
                Common.print_error("Contaminant file " + self.known_contaminant+" does not exist")

    # Calls method to generate cw and quality summary
    def cw_quality_summary(self):
        print("Collating quality and cumulative summary for the experiment....")
        cw_qualitysummary.cw_qualitysummary(self.output_directory, "cw_summary.txt", "quality_summary.txt")

    # Calls method to generate tall skinny file
    def tall_skinny(self):

        if not self.test_mode:
            print("Generating tall-skinny barcode fractions....")
            tall_skinny.tall_skinny(self.output_directory, "tall_skinny.txt")
        else:
            print("Count and Fraction matrix would be generated")

    # Call method to create barcode matrix
    def matricize(self):

        if not self.test_mode:
            print("Generating count and fraction matrix....")
            matricize.matricize(self.output_directory, "matrix_count.txt", "matrix_fraction.txt")
        else:
            print("Count and Fraction matrix would be generated")

    # Go over all the fastq files in the directory and count the barcodes
    def count_barcodes_on_samples(self):

        # Iterate over all the samples in the input directory
        for sample in os.listdir(self.input_directory):
            if sample.startswith('.'):
                continue
            # Check for regex if any
            if self.sample_pattern_regex != "":
                if not re.match(self.sample_pattern_regex, sample):
                    continue

            sample_path = self.input_directory+"/"+sample

            clt_object = ClonTracerCountBarcodes(input_file_list=list(),
                                                 output_prefix=self.output_directory+"/"+sample,
                                                 barcode_index=self.barcode_index,
                                                 phred_format=self.phred_format,
                                                 known_contaminant=self.known_contaminant,
                                                 library_mode=self.library_mode)
            for files in os.listdir(sample_path):
                if "fastq" in files or "fq" in files:
                    clt_object.input_file_list.append(sample_path+"/"+files)

            clt_object.input_files = ','.join(clt_object.input_file_list)

            clt_object.count_barcodes()
            clt_object.simplify_barcodes()


